unit Character;

interface

uses
  World, Graphics, Windows;

const
  GroundVelIncrement: Extended = 0.5; //Inkrement za Vx trcanje
  AirVelIncrement: Extended = 0.125; //Inkrement za Vx letenje
  WaterVelIncrement: Extended = 0.06125; //Inkrement za Vx ronjenje

  GroundSmallMaxVel: Extended = 4; //Max Vx kada hoda
  GroundLargeMaxVel: Extended = 7.5; //Max Vx kada trci
  WaterMaxVel: Extended = 2;//Max Vx kada roni
  AirMaxVel: Extended = 9; //Max Vx kada leti

  GroundTurnDelay: Extended = 3;
  WaterTurnDelay: Extended = 2;
  AirTurnDelay: Extended = 3;

  GroundFriction: Extended = 0.25; //Trenje kada stane sa trcanjem
  AirFriction: Extended = 0.0625; //Trenje kada stane sa letenjem
  WaterFriction: Extended = 0.03125; //Trenje kada stane sa ronjenjem

  GroundSmallJumpImpulse: Extended = 14; //Vertikalni impuls malog skoka
  GroundLargeJumpImpulse: Extended = 16; //Vertikalni impuls velikog skoka
  WaterJumpImpulse: Extended = 3; //Vertikalni impuls u vodi
  AirJumpImpulse: Extended = 10; //Vertikalni impuls u vazduhu

  GroundGravity: Extended = 0.5; //Gravitacija na zemlji
  AirGravity: Extended = 0.125; //Gravitacija letenja
  WaterGravity: Extended = 0.03125; //Gravitacija ronjenja
type
  TCharacter = class
  private
    procedure ParamCheck;
  public
    World: TWorld;
    Gravity: Extended;
    Friction: Extended;
    VelIncrement: Extended;
    MaxVel: Extended;
    TurnDelay: Extended;
    JumpImpulse: Extended;
    //Flegovi (indikatori stanja)
    Flags: set of (flgGoingLeft, flgGoingRight, flgJumping, flgRunning, flgFalling, flgFacingLeft, flgFacingRight, flgWater, flgAir, flgGround);
    X: Extended; //X Koordinata
    Y: Extended; //Y Koordinata
    Width: Integer;  //Sirina igraca, MOZE se bezbedno menjati u toku igre!!!
    Height: Integer; //Visina igraca, MOZE se bezbedno menjati u toku igre (naravno, ako na tren. poziciji ima mesta za novu velicinu)!!!
    PriorX: Extended; //Prethodni (pouzdano "dobri") X
    PriorY: Extended;  //Prethodni (pouzdano "dobri") Y
    Vx: Extended;//X komponenta brzine
    Vy: Extended;//Y komponenta brzine
    Orient: String;
    constructor Create(AWorld: TWorld; AWidth, AHeight, StartX, StartY: Integer);
    procedure Tick; //Otkucaj, ciklus...
    procedure GoLeft; //Idi levo (pritisnut taster)
    procedure StopLeft; //Stani levo (pusten taster)
    procedure GoRight; //Idi desno (pritisnut taster)
    procedure StopRight; //Stani desno (pusten taster)
    procedure GoJump; //Skoci (pritisnut taster)
    procedure StopJump; //Stani sa skokom (pusten taster)
    procedure GoRun; //Trci (pritisnut taster)
    procedure StopRun; //Stani sa trcanjem (pusten taster)
    function GetX: Integer; //Vraca trenutni X
    function GetY: Integer; //Vraca trenutni Y
  end;

implementation

{ TCharacter }

constructor TCharacter.Create(AWorld: TWorld; AWidth, AHeight, StartX, StartY: Integer);
begin
  World := AWorld; //Potrebno zbog World.IsAvailable metoda
  X := StartX; //PocetniX
  Y := StartY; //PocetniY
  Width := AWidth;
  Height := AHeight;
  Vx := 0;
  Vy := 0;
  Flags := [flgFacingRight, flgGround];
  Orient := 'R'; //bice kasnije izbaceno kompletno Orient polje
  ParamCheck;
end;

procedure TCharacter.GoLeft;
begin
  Flags := Flags + [flgGoingLeft];
  Flags := Flags - [flgGoingRight];
  Flags := Flags + [flgFacingLeft];
  Flags := Flags - [flgFacingRight];
  Orient := 'L';
  if not (flgFalling in Flags) then
    if Vx > TurnDelay then  //Da bi SAMO MALO "proklizavao" ako je prethodno isao desno, a pritom hoda
      Vx := TurnDelay;
end;

procedure TCharacter.GoRight;
begin
  Flags := Flags - [flgGoingLeft];
  Flags := Flags + [flgGoingRight];
  Flags := Flags - [flgFacingLeft];
  Flags := Flags + [flgFacingRight];
  Orient := 'R';
  if not (flgFalling in Flags) then
    if Vx < -TurnDelay then  //Da bi SAMO MALO "proklizavao" ako je prethodno isao levo, a pritom hoda
      Vx := -TurnDelay;
end;

procedure TCharacter.GoJump;
begin
  ParamCheck;
  if not ((flgGround in Flags) and (flgFalling in Flags)) then
  begin
    Vy := JumpImpulse;
    Flags := Flags + [flgJumping];
  end;
end;

procedure TCharacter.GoRun;
begin
  Flags := Flags + [flgRunning];
end;

procedure TCharacter.StopJump;
begin
  if flgJumping in Flags then
  begin
    Flags := Flags - [flgJumping];
    if Vy > 6 then  //Da ne bi TRENUTNO stao sa skokom pri pustanju Jump tastera
      Vy := 6;
  end;
end;

procedure TCharacter.StopLeft;
begin
  Flags := Flags - [flgGoingLeft];
end;

procedure TCharacter.StopRight;
begin
  Flags := Flags - [flgGoingRight];
end;

procedure TCharacter.StopRun;
begin
  Flags := Flags - [flgRunning];
end;

procedure TCharacter.Tick;
begin
//Osnovne Provere
  ParamCheck;

//Izracunavanje Vx i X
  if flgGoingRight in Flags then
  begin
    if Vx < MaxVel then
      Vx := Vx + VelIncrement;
    if Vx > MaxVel then
      Vx := MaxVel;
  end;
  if flgGoingLeft in Flags then
  begin
    if Vx > -MaxVel then
      Vx := Vx - VelIncrement;
    if Vx < -MaxVel then
      Vx := -MaxVel;
  end;
  //Ako ne ide ni levo ni desno onda deluje trenje
  if not ((flgGoingLeft in Flags) or (flgGoingRight in Flags)) then
    if Vx <> 0 then
      if Vx > 0 then
        Vx := Vx - Friction
      else
        Vx := Vx + Friction;

  PriorX := X; //Zapamti stari (dobri) X
  X := X + Vx; //Dodaj sracunatu brzinu

  //Proveri i eventualno ispravi novi polozaj
  if not (World.IsAvailablePos(Trunc(X), Trunc(Y)) and World.IsAvailablePos(Trunc(X) + Width, Trunc(Y) + Height) and World.IsAvailablePos(Trunc(X) + Width, Trunc(Y)) and World.IsAvailablePos(Trunc(X), Trunc(Y) + Height)) then
  begin
  //znaci ne valja X
    repeat
      if X < PriorX then
        X := X + 1
      else
        X := X - 1;
    until (World.IsAvailablePos(Trunc(X), Trunc(Y)) and World.IsAvailablePos(Trunc(X) + Width, Trunc(Y) + Height) and World.IsAvailablePos(Trunc(X) + Width, Trunc(Y)) and World.IsAvailablePos(Trunc(X), Trunc(Y) + Height));
    Vx := 0; //Postavi Vx na 0 da se ne bi igrac privremeno "zalepio" za objekat ako se zakuco maksimalnom brzinom i treba vremena da deluje trenje
  end;


//Izracunavanje Vy i Y...
  //Ako je dosao do vrha svog skoka
  if (flgJumping in Flags) and (Vy <= 0) then
  begin
    Flags := Flags - [flgJumping];
    Vy := 0;
  end;

  //Utvrditi da li lik pada, tj. da li je u vazduhu
  //Ako pada onda deluje gravitacija
  if flgFalling in Flags then
    Vy := Vy - Gravity;
  PriorY := Y; //Zapamti stari dobri Y
  Y := Y + Vy;

  //Proveri novi Y
  if not (World.IsAvailablePos(Trunc(X), Trunc(Y)) and World.IsAvailablePos(Trunc(X) + Width, Trunc(Y) + Height) and World.IsAvailablePos(Trunc(X) + Width, Trunc(Y)) and World.IsAvailablePos(Trunc(X), Trunc(Y) + Height)) then
  //znaci ne valja Y
  begin
    repeat
      if Y < PriorY then
        Y := Y + 1
      else
        Y := Y - 1;
    until (World.IsAvailablePos(Trunc(X), Trunc(Y)) and World.IsAvailablePos(Trunc(X) + Width, Trunc(Y) + Height) and World.IsAvailablePos(Trunc(X) + Width, Trunc(Y)) and World.IsAvailablePos(Trunc(X), Trunc(Y) + Height));
    Vy := 0; //Isto ko Vx = 0;
  end;
end;

function TCharacter.GetX: Integer;
begin
  Result := Trunc(X);
end;

function TCharacter.GetY: Integer;
begin
  Result := Trunc(Y);
end;

procedure TCharacter.ParamCheck;
begin
  if World.IsAvailablePos(Trunc(X), Trunc(Y-1)) and World.IsAvailablePos(Trunc(X) + Width, Trunc(Y-1)) then
    Flags := Flags + [flgFalling]
  else
    Flags := Flags - [flgFalling];

  if flgGround in Flags then
  begin
    Gravity := GroundGravity;
    Friction := GroundFriction;
    VelIncrement := GroundVelIncrement;
    if (flgRunning in Flags) and (Abs(Vx) > 0) then
    begin
      MaxVel := GroundLargeMaxVel;
      JumpImpulse := GroundLargeJumpImpulse;
    end
    else
    begin
      MaxVel := GroundSmallMaxVel;
      JumpImpulse := GroundSmallJumpImpulse;
    end;
    TurnDelay := GroundTurnDelay;
  end;

  if flgWater in Flags then
  begin
    Gravity := WaterGravity;
    Friction := WaterFriction;
    VelIncrement := WaterVelIncrement;
    MaxVel := WaterMaxVel;
    JumpImpulse := WaterJumpImpulse;
    TurnDelay := WaterTurnDelay;
  end;

  if flgAir in Flags then
  begin
    Gravity := AirGravity;
    Friction := AirFriction;
    VelIncrement := AirVelIncrement;
    MaxVel := AirMaxVel;
    JumpImpulse := AirJumpImpulse;    
    TurnDelay := AirTurnDelay;
  end;
end;

end.
